#!/usr/local/bin/python
import xmmsclient
import gtk, gtk.glade, gobject
import ConfigParser, os, sys
from Shellac import guiwidgets, xmmsglib, settings, functions, format
import urllib

class Player(gobject.GObject):
    __gproperties__ = {
        "metadata" : (gobject.TYPE_PYOBJECT,
                "metadata map",
                "all the metadata for the currently playing song",
                gobject.PARAM_READWRITE),

        "position" : (gobject.TYPE_INT,
                "current position",
                "current position in the playlist",
                -1,
                sys.maxint,
                -1,
                gobject.PARAM_READWRITE),

        "status" : (gobject.TYPE_INT,
                "current status",
                "the state xmms2d is in - playing, paused or stopped",
                xmmsclient.PLAYBACK_STATUS_STOP,
                xmmsclient.PLAYBACK_STATUS_PAUSE,
                xmmsclient.PLAYBACK_STATUS_STOP,
                gobject.PARAM_READWRITE),

        "playtime" : (gobject.TYPE_INT,
                "current time",
                "elapsed time of the currently playing song",
                0,
                sys.maxint,
                0,
                gobject.PARAM_READWRITE),

        "duration" : (gobject.TYPE_INT,
                "duration",
                "duration of the currently playing song",
                0,
                sys.maxint,
                0,
                gobject.PARAM_READWRITE),

        "id" : (gobject.TYPE_INT,
                "current id",
                "the medialib id of the currently playing item",
                0,
                sys.maxint,
                0,
                gobject.PARAM_READWRITE),

        "volume" : (gobject.TYPE_PYOBJECT,
                "the current volume",
                "the current volume",
                gobject.PARAM_READWRITE)
    }
    __gsignals__ = {
        "playlist-changed" : (gobject.SIGNAL_RUN_LAST,
                gobject.TYPE_NONE,
                (gobject.TYPE_PYOBJECT,)),
        "medialib-entry-changed" : (gobject.SIGNAL_RUN_LAST,
                gobject.TYPE_NONE,
                (gobject.TYPE_INT,))
    }

    def do_get_property(self, property):
        if property.name == "metadata":
            return self.metadata
        elif property.name == "position":
            return self.position
        elif property.name == "status":
            return self.status
        elif property.name == "playtime":
            return self.playtime
        elif property.name == "duration":
            return self.duration
        elif property.name == "id":
            return self.id
        elif property.name == "volume":
            return self.volume
        else:
            raise AttributeError, "Unknown propert %s" % propert.name

    def do_set_property(self, property, value):
        if property.name == "metadata":
            self.metadata = value
        elif property.name == "position":
            self.position = value
        elif property.name == "status":
            self.status = value
        elif property.name == "playtime":
            self.playtime = value
        elif property.name == "duration":
            self.duration = value
        elif property.name == "id":
            self.id = value
        elif property.name == "volume":
            self.volume = value
        else:
            raise AttributeError, "Unknown propert %s" % propert.name

    def __init__(self):
        # Setup gobject stuff
        gobject.GObject.__init__(self)
        self.metadata = {}
        self.position = -1
        self.status = 0
        self.playtime = 0
        self.duration = 0
        self.id = 0
        self.volume = 0

        # Read shellac specific settings
        self.settings = settings.Settings()

        # Make xmms2d connection
        self.client = xmmsclient.XMMS("Shellac")
        if not self.connect_to_server():
            sys.exit(-1)

        self.config_observer = ConfigObserver(self)

        self.init_gui()

        self.init_accelerators()

        self.init_broadcasts()

        # Initial updates
        self.client.playback_status(self.playback_status_cb)
        self.client.playback_current_id(self.playback_current_id_cb)
        self.client.playlist_current_pos(self.playlist_current_pos_cb)

        self.connect("notify::status", self.on_status_changed)
        self.connect("notify::metadata", self.on_metadata_changed)
        self.connect("notify::id", self.on_id_changed)
        self.connect("medialib-entry-changed", self.on_medialib_entry_changed)

        # Register settings callbacks
        self.settings.add_cb("gui", "status_format", self.new_status_format_cb)
        self.settings.add_cb("gui", "show_statusbar", self.show_statusbar_cb)

        # Initial updates
        self.settings.emit("gui", "status_format")
        self.settings.emit("gui", "show_statusbar")

        # Restore old settings
        self.restore_settings()

        # Show everything
        self.window.show_all()

    def init_gui(self):
        glade_file = functions.glade_file("gui")
        self.gxml_menu = gtk.glade.XML(glade_file, "tab_config_menu")
        self.gxml_menu.signal_autoconnect(self)

        self.gxml = gtk.glade.XML(glade_file, "window")
        self.gxml.signal_autoconnect(self)
        w = self.gxml.get_widget

        w("vbox").pack_start(GuiRow(self, "row_1"), False, True, 0)
        w("vbox").pack_start(GuiRow(self, "row_2"), False, True, 0)

        self.window             = w("window")
        self.expander           = w("expander")
        self.expander_label     = w("expander_label")
        self.expander_eventbox  = w("expander_eventbox")
        self.expander_vp        = w("expander_vp")
        self.expander_vbox      = w("expander_vbox")
        self.notebook           = w("notebook")

        # This is required to enable panning of the status string.  Otherwise
        # the 'expander' steals all motion notify events.
        self.expander.add_events(gtk.gdk.BUTTON_MOTION_MASK);

        # Setup drag and drop
        targets = [('text/uri-list', 0, 0), ('text/plain', 0, 0)]
        self.window.drag_dest_set(gtk.DEST_DEFAULT_MOTION | gtk.DEST_DEFAULT_HIGHLIGHT | gtk.DEST_DEFAULT_DROP, targets, gtk.gdk.ACTION_COPY)
        self.window.connect("drag-data-received", self.drag_data_cb)

    def init_accelerators(self):
        accel = gtk.AccelGroup()
        self.gxml_menu.get_widget("preferences").add_accelerator("activate", accel, ord("p"), gtk.gdk.CONTROL_MASK, gtk.ACCEL_VISIBLE)
        self.gxml_menu.get_widget("close").add_accelerator("activate", accel, ord("w"), gtk.gdk.CONTROL_MASK, gtk.ACCEL_VISIBLE)
        self.gxml_menu.get_widget("move_left").add_accelerator("activate", accel, gtk.gdk.keyval_from_name("Left"), gtk.gdk.CONTROL_MASK, gtk.ACCEL_VISIBLE)
        self.gxml_menu.get_widget("move_right").add_accelerator("activate", accel, gtk.gdk.keyval_from_name("Right"), gtk.gdk.CONTROL_MASK, gtk.ACCEL_VISIBLE)

        self.window.add_accel_group(accel)

    def init_broadcasts(self):
        # Register broadcast and signal callbacks
        c = self.client

        c.broadcast_playlist_current_pos(self.playlist_current_pos_cb)
        c.broadcast_playback_status(self.playback_status_cb)
        c.signal_playback_playtime(self.playback_playtime_cb)
        c.broadcast_playback_current_id(self.playback_current_id_cb)
        c.broadcast_playback_volume_changed(self.playback_volume_cb)
        c.broadcast_playlist_changed(self.playlist_changed_cb)
        c.broadcast_medialib_entry_changed(self.medialib_entry_changed_cb)

    def playlist_current_pos_cb(self, res):
        value = res.value()
        if type(value) == int:
            self.set_property("position", value)
        else:
            self.set_property("position", 0)

    def playback_status_cb(self, res):
        value = res.value()
        self.set_property("status", value)

    def playback_playtime_cb(self, res):
        value = res.value()
        self.set_property("playtime", value)
        res.restart()

    def playback_current_id_cb(self, res):
        value = res.value()
        self.set_property("id", value)

    def playback_volume_cb(self, res):
        value = res.value()
        self.set_property("volume", value)


    def playlist_changed_cb(self, res):
        value = res.value()
        self.emit("playlist-changed", value)

    def medialib_entry_changed_cb(self, res):
        id = res.value()
        self.emit("medialib-entry-changed", id)

    def drag_data_cb(self, widget, context, x, y, selection, info, time):
        to_add = []


        for uri in selection.get_uris():
            uri = urllib.url2pathname(uri)

            # Different programs use different conventions on how to encode the
            # uris.
            if uri.startswith("file://localhost"):
                # E.g. Rox
                uri = uri[len("file://localhost"):]
            elif uri.startswith("file:///"):
                # E.g Nautilus
                uri = uri[len("file://"):]
            elif uri.startswith("file://"):
                # E.g. Opera
                uri = uri[len("file:/"):]
            elif uri.startswith("file:/"):
                # E.g. KDE programs
                uri = uri[len("file:"):]


            if os.path.isdir(uri):
                # Add recursively
                for dirpath, dirnames, filenames in os.walk(uri):
                    for filename in filenames:
                        to_add.append("file://" + os.path.join(dirpath, filename))
            else:
                to_add.append("file://" + uri)

        # Firefox transmitts dnd urls as 'text/plain' NOT as 'text/uri-list'
        text = selection.get_text()
        try:
            text = urllib.url2pathname(text)
            if text.startswith("file://"):
                to_add.append(text)
            elif text.startswith("http://"):
                to_add.append(text)
        except AttributeError:
            pass


        to_add.sort()

        # Add them
        for path in to_add:
            self.client.playlist_add(path)

    def connect_to_server(self):
        save_ipcpath = False
        try:
            ipcpath = os.environ["XMMS_PATH"]
        except KeyError:
            ipcpath = self.settings.get("global", "ipcpath")

        while 1:
            try:
                self.client.connect(ipcpath, self.disconnect_cb)
                xmmsglib.GLibConnector(self.client)

                if save_ipcpath:
                    # Save the working ipc path
                    self.settings.set("global", "ipcpath", ipcpath)
                return True
            except IOError:
                glade_file = functions.glade_file("connect")
                gxml = gtk.glade.XML(glade_file)
                gxml.signal_autoconnect(self)

                w = gxml.get_widget
                result = w("connect_dialog").run()

                if result == gtk.RESPONSE_OK:
                    save_ipcpath = w("save").get_active()
                    ipcpath = w("daemon_path").get_text()
                    w("connect_dialog").destroy()
                elif result == gtk.RESPONSE_APPLY:
                    cmd = w("daemon_cmd").get_text()
                    w("connect_dialog").destroy()
                    os.system(cmd)
                else:
                    w("connect_dialog").destroy()
                    return False

    def tab_factory(self, type):
        w = None
        title = None
        if type == "Playlist":
            from Shellac import playlist
            w = playlist.Playlist(self)
            title = "_Playlist"

        elif type == "Playlists":
            from Shellac import playlists
            w = playlists.Playlists(self)
            title = "Playli_sts"

        elif type == "Find":
            from Shellac import find
            w = find.Find(self)
            title = "_Find"

        elif type == "Library":
            from Shellac import library
            w = library.Library(self)
            title = "_Library"

        elif type == "Equalizer":
            from Shellac import equalizer
            w = equalizer.Equalizer(self)
            title = "E_qualizer"
        
        elif type == "Stream browser":
            from Shellac import streambrowser
            w = streambrowser.StreamBrowser(self)
            title = "Stream browser"

        elif type == "File browser":
            from Shellac import filebrowser
            w = filebrowser.FileBrowser(self)
            title = "File browser"
        return (w, title)

    def add_tab(self, type):
        """
        Add a tab of the specified type.
        Nothing prevents several tabs of the same type.
        """
        # Construct the right tab...
        w, title = self.tab_factory(type)
        
        # ...and add it
        if w != None:
            label = gtk.Label(title)
            label.set_use_underline(True)
            self.notebook.append_page(w, label)
            self.notebook.show_all()

    def add_detached(self, name):
        """
        Construct a new top level window showing the specified tab type.
        """
        widget, title = self.tab_factory(name)

        # Create the new window
        if widget != None:
            x, y, w, h = self.settings.get("windows", name.lower()).split(",")
            win = gtk.Window()
            win.set_title(name)
            win.resize(int(w), int(h))
            win.set_size_request(1, -1)
            win.move(int(x), int(y))
            win.add(widget)
            win.show_all()
            win.connect("delete-event", self.on_detached_delete)

    def set_status_string(self):
        if self.get_property("status") != xmmsclient.PLAYBACK_STATUS_STOP:
            text = self.format.format(self.get_property("metadata"))
        else:
            text = "Stopped"

        self.expander_label.set_markup(text)
        self.expander.queue_draw()

    def save_current_info(self, id):
        """
        Cache the metadata info about an id.
        """
        if id > 0:
            self.client.medialib_get_info(id, self.get_metadata_cb)

    def get_metadata_cb(self, res):
        metadata = res.value()

        if metadata.has_key(("server", "duration")):
            self.set_property("duration", max (int(metadata[("server", "duration")]), 0))
        self.set_property("metadata", metadata)
        
    def save_settings(self):
        # Save window position
        x, y = self.window.get_position()
        self.settings.set("windows", "main_pos", "%d,%d" % (x, y))

        # Save window size
        if self.expander.get_expanded() == True:
            w, h = self.window.get_size()
            self.settings.set("windows", "main_size", "%d,%d" % (w, h))

        # Save expanded state
        state = str(int(self.expander.get_expanded()))
        self.settings.set("gui", "expanded", state)

        # Save visible tabs
        nb = self.notebook
        tabs = ",".join([nb.get_tab_label_text(x)
                for x in nb.get_children()])
        self.settings.set("gui", "tabs", tabs)

        # Save active tab
        page = str(nb.get_current_page())
        self.settings.set("gui", "active_tab", page)

        # Save detached windows
        wins = [w for w in gtk.window_list_toplevels() if w.get_property("type") == gtk.WINDOW_TOPLEVEL]
        for win in wins:
            x, y = win.get_position()
            w, h = win.get_size()
            title = win.get_title()
            section = "windows"
            option = title.lower()
            value = "%d,%d,%d,%d" % (x, y, w, h)
            self.settings.set(section, option, value)

        TABS = ["Playlist", "Library", "Find", "Playlists", "Equalizer", "Stream browser", "File browser"]
        visible = ",".join([win.get_title() for win in wins if win.get_title() in TABS])
        self.settings.set("windows", "visible", visible)

        # Save column order/width
        for tab in nb.get_children():
            try:
                value = ",".join(["%s:%d" % (col.get_title(), max(1, col.get_width()))
                        for col in tab.song_view.view.get_columns()])
                section = tab.song_view.section
                self.settings.set(section, "columns", value)
            except:
                pass

        # Save to disk
        self.settings.write()

    def restore_settings(self):
        # Restore window position
        x, y = self.settings.get("windows", "main_pos").split(",")
        self.window.move(int(x), int(y))

        # Restore window size
        w, h = self.settings.get("windows", "main_size").split(",")
        self.window.resize(int(w), int(h))

        # Restore expanded state
        expanded = bool(int(self.settings.get("gui", "expanded")))
        if not expanded:
            self.expander.activate()

        # Restore visible tabs
        tabs = self.settings.get("gui", "tabs")
        for tab in tabs.split(","):
            self.add_tab(tab)

        # Restore active tab
        active_tab = int(self.settings.get("gui", "active_tab"))
        if active_tab >= 0:
            self.notebook.set_current_page(active_tab)

        # Restore detached tabs
        wins = self.settings.get("windows", "visible")
        for win in wins.split(","):
            self.add_detached(win)



    ################
    # Event handlers
    ################
    def on_expander_button_release(self, widget, event):
        if event.button == 2:
            self.expander_eventbox.grab_remove()

    def on_expander_button_press(self, widget, event):
        """
        Using mouse button 2 saves the click position and adjustment
        position required to pan a long status title.
        """
        if event.button == 2:
            self.pan_begin_x = event.x_root
            self.pan_begin_hadj = self.expander_vp.get_hadjustment().get_value()
            self.expander_eventbox.grab_add()
        elif event.button == 3:
            self.gxml_menu.get_widget("tab_config_menu").popup(
                    None, None, None, event.button, event.time)


    def on_expander_activated(self, expander):
        """
        This is executed before the expander has actually changed state.
        We must hide or show the contents before the state change.
        When going from exapnded to non-expanded mode, we must save the
        original window size. Else the correct expanded size can't be
        restored at startup in case we close the program in non-expanded
        mode.
        """
        if self.expander.get_expanded():
            w, h = self.window.get_size()
            self.settings.set("windows", "main_size", "%d,%d" % (w, h))
            self.expander.child.hide_all()
        else:
            self.expander.child.show_all()

    def on_after_expander_activated(self, expander):
        """
        After the expanded status change, we must set the correct size.
        """
        width, height = self.window.get_size()
        if self.expander.get_expanded() == False:
            self.window.resize(width, 1)
        else:
            w, h = self.settings.get("windows", "main_size").split(",")
            height = int(h)
            self.window.resize(width, height)

    def on_expander_eventbox_motion_notify(self, widget, event):
        """
        Scroll the status string depending on how far we have dragged/panned
        from the start position.
        """
        #TODO: Check that button 2 is pressed
        dx = event.x_root - self.pan_begin_x
        hadj = self.expander_vp.get_hadjustment()
        new_hadj = min(self.pan_begin_hadj - dx, hadj.upper - hadj.page_size)
        hadj.set_value(new_hadj)

    def on_delete_event(self, widget, event):
        """
        When quitting the program, some configuration variables must be
        saved.
        """
        self.save_settings()

        gtk.main_quit()


    def on_key_press(self, widget, event):
        """
        Capture global shortcut keys.
        """

        # Quit [Escape]
        if event.keyval == gtk.gdk.keyval_from_name("Escape"):
            self.save_settings()
            gtk.main_quit()

    ################
    # Menu callbacks
    ################
    def on_move_left_activate(self, widget):
        current_page = self.notebook.get_current_page()
        current_widget = self.notebook.get_nth_page(current_page)
        new_pos = max(0, current_page - 1)
        self.notebook.reorder_child(current_widget, new_pos)

    def on_move_right_activate(self, widget):
        current_page = self.notebook.get_current_page()
        current_widget = self.notebook.get_nth_page(current_page)
        new_pos = min(current_page + 1, self.notebook.get_n_pages())
        self.notebook.reorder_child(current_widget, new_pos)

    def on_new_tab(self, widget):
        name = widget.get_child().get_text()
        self.add_tab(name)

    def on_new_detached(self, widget):
        name = widget.get_child().get_text()
        self.add_detached(name)

    def on_preferences_activate(self, widget):
        from Shellac import settingseditor
        settingseditor.SettingsEditor(self)

    def on_detached_delete(self, widget, event):
        """
        When a detached window is deleted, save its size and position.
        """
        widget = widget.get_child()
        try:
            widget.destructor()
        except AttributeError:
            # Nothing needs to be cleaned up.
            pass

    def on_close_activate(self, widget):
        """
        Remove the closed tab.
        """
        current_page = self.notebook.get_current_page()
        widget = self.notebook.get_nth_page(current_page)
        try:
            widget.destructor()
        except AttributeError:
            # Nothing needs to be cleaned up.
            pass
        self.notebook.remove_page(current_page)

    ###########
    # Callbacks
    ###########
    def on_metadata_changed(self, obj, property):
        self.set_status_string()

    def on_status_changed(self, obj, property):
        self.set_status_string()

    def on_id_changed(self, obj, property):
        id = self.get_property("id")
        self.save_current_info(id)

    def show_statusbar_cb(self, option, section, value):
        
        if bool(int(value)):
            statusbar = guiwidgets.Statusbar(self)
            self.expander_vbox.pack_start(statusbar, False, False)
            self.expander_vbox.show_all()
        else:
            for child in self.expander_vbox.get_children():
                if type(child) == guiwidgets.Statusbar:
                    child.destructor()
                    self.expander_vbox.remove(child)

    def disconnect_cb(self, value):
        print "Disconnecting. Daemon shut down."
        gtk.main_quit()

    def on_medialib_entry_changed(self, obj, id):
        if self.get_property("metadata").get(("server", "id")) == id:
            self.save_current_info(id)

    def new_status_format_cb(self, section, option, value):
        self.format = format.Format(value)
        self.set_status_string()


class GuiRow(gtk.HBox):
    def __init__(self, player, row_name):
        gtk.HBox.__init__(self)
        self.player = player
        self.row_name = row_name
        player.settings.add_cb("gui", row_name, self.gui_row_changed_cb)
        player.settings.emit("gui", row_name)

    def gui_row_changed_cb(self, section, option, value):
        # Remove all old widgets
        for child in self.get_children():
            try:
                child.destructor()
            except AttributeError, e:
                # Nothing needs to be cleaned up.
                pass
            self.remove(child)

        # Add the new ones
        for name in value.split(","):
            guiwidgets.add_gui_widget(self.player, name, self)

        self.show_all()

class ConfigObserver:
    """
    Class managing subscriptions to xmms2d configuration values.
    Use this to execute something when a specific configuration value has
    changed.
    """
    def __init__(self, player):
        # The callbacks are stored in a map between keys and a list of callback
        # functions.
        self.__callbacks = {}
        self.player = player
        player.client.broadcast_configval_changed(self.__callback)

    def add(self, key, callback):
        """
        Subscribe to 'key' and run 'callback' upon changes.
        """
        if self.__callbacks.has_key(key):
            self.__callbacks[key].append(callback)
        else:
            self.__callbacks[key] = [callback]

    def remove(self, key, callback):
        """
        Unregister 'callback' from 'key'.
        """
        for func in self.__callbacks[key]:
            if func == callback:
                self.__callbacks[key].remove(callback)
        

    def emit(self, key):
        """
        Emit a changed signal to everybody subscribing to 'key', even if
        nothing has changed.
        """
        handler = self.EmitHandler(key, self.__callbacks[key])
        self.player.client.configval_get(key, handler)

    def __callback(self, res):
        dict = res.value()
        key = dict.keys()[0]
        value = dict[key]
        try:
            for func in self.__callbacks[key]:
                func(key, value)
        except KeyError:
            # Nobody is subscribing to this key
            pass

    class EmitHandler:
        def __init__(self, key, callbacks):
            self.key = key
            self.callbacks = callbacks

        def __call__(self, res):
            value = res.value()
            try:
                for func in self.callbacks:
                    func(self.key, value)
            except KeyError:
                # Nobody is subscribing to this key
                pass

if __name__ == "__main__":
    gtk.threads_init()
    gtk.threads_enter()
    if gtk.check_version(2, 8, 0):
        gobject.type_register(Player)

    Player()

    gtk.main()
    gtk.threads_leave()
