import gtk, gtk.glade, os, functions, song_tree_view, xmmsclient, urllib, xml.sax.handler, re, threading, sys

class StreamBrowser(gtk.VBox):
    def __init__(self, player):
        gtk.VBox.__init__(self)
        self.player = player
        self.song_view = song_tree_view.SongView(player, "stream_browser")

        glade_file = functions.glade_file("streambrowser")
        self.gxml_menu = gtk.glade.XML(glade_file, "popup")
        self.gxml_menu.signal_autoconnect(self)

        self.song_view.view.connect("row-activated", self.on_row_activated)
        self.song_view.view.connect("button-press-event", self.on_button_press)
        self.song_view.view.connect("button-release-event", self.on_button_release)
        self.song_view.view.connect("row-expanded", self.on_row_expanded)

        self.model = gtk.TreeStore(object)
        self.song_view.view.set_model(self.model)

        self.pack_start(self.song_view)

        # Initial update
        self.reload()

    def reload(self):
        """
        Forces a reload
        """
        self.model.clear()
        plugins = ["Shoutcast", "Xiph"]
        for plugin in plugins:
            dict = {"plugin" : plugin}
            iter = self.model.append(None, (dict,))
            self.model.append(iter, (None,))

    def show_menu(self, button=0, time=0):
        """
        Popup the context menu.
        """
        w = self.gxml_menu.get_widget
        selected = self.song_view.view.get_selection().count_selected_rows() > 0
        w("play").set_sensitive(selected)
        w("add").set_sensitive(selected)
        w("popup").popup(None, None, None, button, time)

    def destructor(self):
        self.song_view.destructor()

    ################
    # Event handlers
    ################
    def on_row_activated(self, treeview, path, view_column):
        """
        Double clicking a row is the same as choosing 'Play' from the context
        menu.
        """
        self.gxml_menu.get_widget("play").activate()

    def on_button_press(self, widget, event):
        """
        Prevend deselection of multiple rows at mouse button press.
        """
        if event.button == 3:
            return self.song_view.view.get_selection().count_selected_rows() > 1

    def on_button_release(self, widget, event):
        """
        Show the context menu.
        """
        if event.button == 3:
            self.show_menu(event.button, event.time)

    def on_row_expanded(self, view, iter, path):
        model = view.get_model()
        child = model.iter_children(iter)
        dict = model.get_value(iter, 0)
        if model.get_value(child, 0) is None:
            model.remove(child)

            if len(path) == 1:
                plugin = plugin_factory(dict["plugin"], self, path)
                plugin.start()

            elif len(path) == 2:
                f = urllib.urlopen(dict["url"])
                pattern = re.compile("^File[0-9]+=.+")
                for line in f.readlines():
                    if pattern.match(line) != None:
                        url = line.split("=")[-1].strip()

                        self.model.append(iter, ({"url" : url},))

        #view.expand_row(path, False)
    
    ################
    # Menu callbacks
    ################
    def on_play_activate(self, widget):
        """
        Replace current playlist with selected playlists.
        """
        self.player.client.playlist_clear()
        self.gxml_menu.get_widget("add").activate()
        self.player.client.playlist_set_next(0)
        self.player.client.playback_tickle()
        self.player.client.playback_start()

    def on_add_activate(self, widget):
        """
        Add all selected streams to the playlist.
        """
        model, paths = self.song_view.view.get_selection().get_selected_rows()

        for path in paths:
            if len(path) == 2:
                iter = model.get_iter(path)
                if model.iter_has_child(iter):
                    view = self.song_view.view
                    view.expand_row(path, False)

                    path = path + (0,)
                    column = view.get_column(0)
                url = self.model[path][0]["url"]
                self.player.client.playlist_add(url)

            elif len(path) == 3:
                url = self.model[path][0]["url"]
                self.player.client.playlist_add(url)

    def on_refresh_activate(self, widget):
        """
        Update the playlists list.
        """
        self.reload()

def plugin_factory(name, caller, path):
    if name == "Shoutcast":
        return ShoutcastPlugin(caller, path)
    elif name == "Xiph":
        return XiphPlugin(caller, path)

class ShoutcastPlugin(threading.Thread):
    class ShoutcastHandler(xml.sax.handler.ContentHandler):
        def __init__(self):
            self.list = []
            self.info = {}
            self.current_field = None

            self.metadata_map = {
                "Name" : "channel",
                "Genre" : "genre",
                "Nowplaying" : "nowplaying",
                "Listeners" : "listeners",
                "Bitrate" : "bitrate"
            }

        def startElement(self, name, attributes):
            if name == "entry":
                self.info = {}
                self.info["url"] = attributes["Playstring"]
            elif name in self.metadata_map.keys():
                self.current_field = name

        def characters(self, data):
            if self.current_field != None and data != "":
                self.info[self.metadata_map[self.current_field]] = data

        def endElement(self, name):
            self.current_field = None
            if name == "entry":
                self.list.append(self.info)

    def __init__(self, caller, path):
        threading.Thread.__init__(self)
        self.caller = caller
        self.path = path

    def run(self):
        url =" http://www.shoutcast.com/sbin/xmllister.phtml?service=shellac&no_compress=1&limit=100" 
        parser = xml.sax.make_parser()
        handler = self.ShoutcastHandler()
        parser.setContentHandler(handler)
        parser.parse(url)

        iter = self.caller.model.get_iter(self.path)

        for info in handler.list:
            new_iter = self.caller.model.append(iter, (info,))
            self.caller.model.append(new_iter, (None,))
        self.caller.song_view.view.expand_row(self.path, False)
        self.caller.song_view.view.queue_draw()

class XiphPlugin(threading.Thread):
    class XiphHandler(xml.sax.handler.ContentHandler):
        def __init__(self):
            self.list = []
            self.info = {}
            self.current_field = None

            self.metadata_map = {
                "server_name" : "channel",
                "listen_url" : "url",
                "bitrate" : "bitrate",
                "genre" : "genre",
                "current_song" : "nowplaying",
                "Listeners" : "listeners",
            }

        def startElement(self, name, attributes):
            if name == "entry":
                self.info = {}
            elif name in self.metadata_map.keys():
                self.current_field = name

        def characters(self, data):
            if self.current_field != None and data != "":
                self.info[self.metadata_map[self.current_field]] = data

        def endElement(self, name):
            self.current_field = None
            if name == "entry":
                self.list.append(self.info)

    def __init__(self, caller, path):
        threading.Thread.__init__(self)
        self.caller = caller
        self.path = path

    def run(self):
        url =" http://dir.xiph.org/yp.xml" 
        parser = xml.sax.make_parser()
        handler = self.XiphHandler()
        parser.setContentHandler(handler)
        try:
            parser.parse(url)

            iter = self.caller.model.get_iter(self.path)

            for info in handler.list:
                new_iter = self.caller.model.append(iter, (info,))
            self.caller.song_view.view.expand_row(self.path, False)
            self.caller.song_view.view.queue_draw()
        except Exception, e:
            print >>sys.stderr, "Streambrowser:", e

